package installer

// package install provides helper functions for all download and installation of following tasks:
// 1. downloading and install nuclei templates
// 2. download and update nuclei binary (i.e self update)
// 3. version check for nuclei binary & templates
